#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# a class for reading (and hopefully eventually writing) DWG files
#

from __future__ import generators

import types

class Dwg:
    """The class for revision-neutral handling of DWG files.

The Dwg class is designed to provide a version-neutral interface
for reading DWG files. A Dwg instance has the following methods:

getHandle(): Get the opened file handle associated with the instance.
{get/set}Version(): Get/Set the version of the underlying DWG file.
{get/set}Offset(): Get/Set the offsets to sections with the DWG file.
getOffsetKeys(): Retrieve the section names for which offsets are stored.
setReader(): Set the function used for reading a particular DWG section.
{get/set}Header(): Get/Set a header variable used in the DWG file.
getHeaderKeys(): Get all the header variables in the DWG file.
{get/set}Class(): Get/Set a class used in the DWG file.
getClassKeys(): Get all the classes used in the DWG file.
{get/set}Object(): Get/Set an entity stored in the DWG file.
{get/set}ImageData(): Get/Set any bitmap image data in the DWG file.

Older versions of the DWG format lack image data and header variables,
and the presence of image data is optional. Also, the header data
variables keys can be different from one release to another.
    """
    def __init__(self, filename):
        """Initialize a Dwg instance.

dwg = Dwg(filename)

The one required argument is the path to a DWG file.
        """
        self.__fp = file(filename)
        self.__version = None
        self.__offsets = {} # file offsets
        self.__entities = [] # entity offset data
        self.__dxfnames = {} # class/dxfname map
        self.__readers = {}
        self.__headers = {}
        self.__classes = {}
        self.__objects = []
        self.__wmfdata = None # not before R13
        self.__bmpdata = None # not before R13
        self.__index = 0
        self.__len = 0
        self.setVersion()

    def __del__(self):
        self.__fp.close()

    def getHandle(self):
        """Return the opened file handle associated with the DWG file.

getHandle()
        """
        return self.__fp

    def setVersion(self):
        """Store the version of the DWG file.

setVersion()
        """
        _fp = self.__fp
        _fp.seek(0, 0)
        _buf = _fp.read(6)
        if _buf == 'AC1009':
            _ver = 'R12'
        elif _buf == 'AC1010': # autocad 10?
            _ver = 'R12+'
        elif _buf == 'AC1012':
            import dwg1314
            dwg1314.initialize_dwg(self)
            _ver = 'R13'
        elif _buf == 'AC1014':
            import dwg1314
            dwg1314.initialize_dwg(self)
            _ver = 'R14'
        elif _buf == 'AC1015':
            import dwg15
            dwg15.initialize_dwg(self)
            _ver = 'R15'
        else:
            _ver = None # unknown file - maybe raise an error?
            self.__fp.close()
        self.__version = _ver

    def getVersion(self):
        """Return the version of the DWG file in use.

getVersion()
        """
        if self.__version is None:
            self.setVersion()
        return self.__version

    def setOffset(self, key, value):
        """Store the offset to a section within the DWG file.

setOffset(key, value)

This method requires two arguments:

key: A text string giving the section name.
value: A tuple of two objects.

Valid keys are one of the following:
HEADERS, CLASSES, OBJECTS, IMAGES, UNKNOWN, R14DATA, R15REC5

The tuple is of the following format:

(offset, size)

offset: Offset in the file
size: The size of the section; This argument can be 'None'
        """
        if not isinstance(key, str):
            raise TypeError, "Invalid offset key: " + str(key)
        if not isinstance(value, tuple):
            raise TypeError, "Invalid offset tuple: " + str(value)
        if not len(value) == 2:
            raise ValueError, "Invalid offset tuple: " + str(value)
        if key in self.__offsets:
            raise ValueError, "Key already set: " + key
        if (key == 'HEADERS' or
            key == 'CLASSES' or
            key == 'OBJECTS' or
            key == 'IMAGES' or
            key == 'UNKNOWN' or
            key == 'R14DATA' or
            key == 'R14REC5'):
            self.__offsets[key] = value
        else:
            raise ValueError, "Unexpected offset key: " + key

    def getOffsetKeys(self):
        """Return the strings giving the sections for which offset data is stored.

getOffsetKeys()

This method returns the keys used in the setOffset() calls. The ordering
of the keys is random. The offset data can be retrieved by calling the
getOffset() method.
        """
        return self.__offsets.keys()

    def getOffset(self, key):
        """Return the section data associated with a particular section key.

getOffset(key)

Argument 'key' should be a string returned from getOffsetKeys().
        """
        return self.__offsets[key]

    def setReader(self, key, value):
        """Store a function which reads a section of the DWG file.

setReader(key, value)

This function has two required arguments:

key: A text string giving the reader type
value: The function used to read the section of the DWG file.

Valid keys are:
HEADERS, CLASSES, OBJECTS, IMAGES
        """
        if not isinstance(key, str):
            raise TypeError, "Invalid offset key: " + str(key)
        if not isinstance(value, types.FunctionType):
            raise TypeError, "Invalid reader for '%s': %s" % (key, str(value))
        if (key == 'HEADERS' or
            key == 'CLASSES' or
            key == 'OBJECTS' or
            key == 'OFFSETS' or
            key == 'OBJECT' or
            key == 'IMAGES'):
            self.__readers[key] = value
        else:
            raise ValueError, "Unexpected reader key: " + key

    def getReader(self, key):
        """Return the function for reading a DWG file section.

getReader(key)        
        """
        return self.__readers[key]
    
    def setHeader(self, key, value):
        """Store a header variable found in the DWG file.

setHeader(key, value)

This method has two arguments:

key: The header variable
value: Its value.

The 'key' must be a string, and the value can be any type of Python
object (string, int, double, tuple, etc...)
        """
        if not isinstance(key, str):
            raise TypeError, "Invalid header key: " + str(key)
        self.__headers[key] = value

    def getHeaderKeys(self):
        """Return the various header variables found in the DWG file.

getHeaderKeys()

This method returns a list of strings, with each being one of the
header variables found in the DWG file. The data associated with
each header can be retrieved with the getHeader() method.
        """
        if not len(self.__headers):
            if 'HEADERS' in self.__readers:
                _reader = self.__readers['HEADERS']
                _reader(self)
        return self.__headers.keys()

    def getHeader(self, key):
        """Return the associated value for a particular header variable.

getHeader(key)

Argument 'key' should be one of the strings returned from getHeaderKeys()
        """
        return self.__headers[key]

    def setDxfName(self, key, dxfname):
        """Store the mapping between the type number and DXF name.

setDxfName(key, dxfname)

Argument 'key' is an integer, and argument 'dxfname' is a string.
This data is found in the class data section in R13, R14, and R15
DWG files.
        """
        if not isinstance(key, int):
            raise TypeError, "Invalid dxfname key: " + str(key)
        if not isinstance(dxfname, str):
            raise TypeError, "Invalid dxfname: " + str(key)
        self.__dxfnames[key] = dxfname

    def getDxfName(self, key):
        """Return the dxfname for a given class key.

getDxfName(key)

Argument 'key' should be an integer. This method returns a
string if there is a class for the key value. Otherwise this
method returns None.

        """
        if not isinstance(key, int):
            raise TypeError, "Invalid dxfname key: " + str(key)
        return self.__dxfnames.get(key)

    def setClass(self, key, value):
        """Store a class variable found in the DWG file.

setClass(key, value)

This method has two required arguments:

key: An integer value
value: A tuple

The contents of the tuple are as follows:
(appname, cplusplusname, dxfname, zombie, id)

The tuple data comes from the R13/R14/R15 spec.
        """
        if not isinstance(key, int):
            raise TypeError, "Non-integer class key: " + str(key)
        if not isinstance(value, tuple):
            raise TypeError, "Non-tuple class value: " + str(value)
        self.__classes[key] = value

    def getClassKeys(self):
        """Return the various classes found in the DWG file.

getClassKeys()

This method returns a list of class values. The data associated
for each class can be obtained with getClass().
        """
        if not len(self.__classes):
            if 'CLASSES' in self.__readers:
                _reader = self.__readers['CLASSES']
                _reader(self)
        return self.__classes.keys()

    def getClass(self, key):
        """Return the class data for a given class.

getClass()

This method returns a tuple holding the class data.
        """
        return self.__classes.get(key)

    def addEntityOffset(self, handle, offset):
        """Store handle/offset data for an entity in the DWG file.

addEntityOffset(handle, offset)
        """
        self.__entities.append((handle, offset))

    def getEntityOffset(self):
        """Return the list of handle/offset data for the DWG file entities.

getEntityOffset()
        """
        for _entdata in self.__entities:
            _id, _offset = _entdata
            yield _offset

    def setObject(self, obj):
        """Store an entity found within the DWG file.

setObject(obj)

Argument 'obj' must be an dwgEntity instance.
        """
        if not isinstance(obj, dwgEntity):
            raise TypeError, "Invalid DWG object: " + str(obj)
        self.__objects.append(obj)

    def getObject(self):
        """Return a single object from the DWG file.

getObject()

This method can be called to extract the dwgEntity objects from
the DWG file one object at a time. The first call gets the
first entity, and each subsequent call retrieves the following
entity.
        """
        if not len(self.__classes):
            if 'CLASSES' in self.__readers:
                _reader = self.__readers['CLASSES']
                _reader(self)
        if not len(self.__entities):
            if 'OFFSETS' in self.__readers:
                _reader = self.__readers['OFFSETS']
                _reader(self)
        if 'OBJECT' not in self.__readers:
            raise StopIteration
        _reader = self.__readers['OBJECT']
        for entdata in self.__entities:
            _id, _offset = entdata
            yield _reader(self, _offset)

    def rewind(self):
        if not len(self.__classes):
            if 'CLASSES' in self.__readers:
                _reader = self.__readers['CLASSES']
                _reader(self)
        if not len(self.__entities):
            if 'OFFSETS' in self.__readers:
                _reader = self.__readers['OFFSETS']
                _reader(self)
        self.__index = 0
        self.__len = len(self.__entities)

    def next_object(self):
        if self.__index == self.__len:
            return
        _id, _offset = self.__entities[self.__index]
        _reader = self.__readers['OBJECT']
        self.__index = self.__index + 1
        return _reader(self, _offset)

    def getEntities(self):
        if not len(self.__classes):
            if 'CLASSES' in self.__readers:
                _reader = self.__readers['CLASSES']
                _reader(self)
        if not len(self.__entities):
            if 'OFFSETS' in self.__readers:
                _reader = self.__readers['OFFSETS']
                _reader(self)
        return self.__entities

    def getObjects(self):
        """Return all the stored objects found in the DWG file.

getObjects()

This method returns a list of dwgEntity objects.
        """
        if not len(self.__classes):
            if 'CLASSES' in self.__readers:
                _reader = self.__readers['CLASSES']
                _reader(self)
        if not len(self.__entities):
            if 'OFFSETS' in self.__readers:
                _reader = self.__readers['OFFSETS']
                _reader(self)
        if not len(self.__objects):
            if 'OBJECTS' in self.__readers:
                _reader = self.__readers['OBJECTS']
                _reader(self)
        return self.__objects[:]

    def setImageData(self, imagetype, data):
        """Store the bitmap image data found in the DWG file.

setImageData(imagetype, data)

This method has two required arguments:

imagetype: A string - either 'BMP' or 'WMF'
data: The image data

The format in which the data is stored is not checked. The R13/R14/R15
readers use array.array instances for this. The image data is not
found in R12 and earlier files.
        """
        if not isinstance(imagetype, str):
            raise TypeError, "Invalid image type: " + str(imagetype)
        if imagetype == 'BMP':
            self.__bmpdata = data
        elif imagetype == 'WMF':
            self.__wmfdata = data
        else:
            raise ValueError, "Unexpected image type: " + imagetype

    def getImageData(self, imagetype):
        """Return the image data found in the DWG file.

getImageData(imagetype)

This method requires a single argument:

imagetype: A string - either 'BMP' or 'WMF'.

There is no image data in R12 and earlier files, and image data
is optional in R13 and later files. If there is no image data
this method returns None.
        """
        if not isinstance(imagetype, str):
            raise TypeError, "Invalid image type: " + str(imagetype)
        if imagetype == 'BMP':
            return self.__bmpdata
        elif imagetype == 'WMF':
            return self.__wmfdata
        else:
            raise ValueError, "Unexpected image type: " + imagetype

class dwgEntity:
    """A generic class for storing information about DWG objects.

The dwgEntity class provides a revision neutral means of storing
data found within the DWG file for all the drawing entities. Some entities
are visible entities like lines, circles, and arcs, and others are
non-graphical entities like tables. The dwgEntity class has the
following methods:

{get/set}Type(): Get/Set the entity type.
{get/set}Handle(): Get/Set the entity handle (identifier).
{get/set}EntityData(): Get/Set some information about a DWG entity.
getEntityKeys(): Return the keys used for storing entity data.

Each different type of DWG entity will have different keys stored,
and the number of keys varies based on the entity and the ability to
decode the information found in the DWG file itself.

DWG entities in R13, R14, and R15 files have a large number of
shared data attributes. The followin method are available for
examining this information. Many of the following methods will only
be useful when reading the entities from the DWG file itself.

{get/set}Version(): Get/Set a DWG version in the entity
{get/set}Mode(): Get/Set the entity mode.
{get/set}NumReactors(): Get/Set the number of reactors.
{get/set}NoLinks(): Get/Set the entity linkage flag.
{get/set}IsLayerByLinetype(): ????
{get/set}Color(): Get/Set the entity color.
{get/set}LinetypeScale(): Get/Set the linetype scale factor.
{get/set}LinetypeFlags(): Get/Set the linetype flags.
{get/set}PlotstyleFlags(): Get/Set the plotstyle flags.
{get/set}Invisibility(): Get/Set the entity invisiblity flags.
{get/set}Lineweight(): Get/Set the entity lineweight factor.
{get/set}Subentity(): Get/Set the subentity flgas.
addReactor(): Add a handle of a reactor object.
getReactors(): Get all the reactors associated with an object.
{get/set}Xdicobj(): ????
{get/set}Layer(): Get/Set the layer where the object is placed.
{get/set}Linetype(): Get/Set the entity linetype
getPrevious(): Get the previous entity in a entity chain.
getNext(): Get the subsequent entity in an entity chain.
{get/set}Plotstyle(): Get/Set the plotstyle flags.

Some of the methods are particular to R13 and R14 files, and some
are particular to R15 files.
    """
    def __init__(self):
        """Initialize a dwgEntity instance.

ent = dwgEntity()

This method requires no arguments.
        """
        self.__cdata = {} # "common" stuff for all entities
        self.__edata = {} # entity specfic data

    def getEntityKeys(self):
        """Return the keys used to store entity specific data.

getEntityKeys()

THis method returns an unsorted list of strings. The value associated
with each key can be obtained by calling the getEntityData() method.
        """
        return self.__edata.keys()

    def setEntityData(self, key, value):
        """Store entity specfic data in the dwgEntity.

setEntityData(key, value)

This method requires two arguments:

key: A string used to describe the stored data.
value: Any Python type.

        """
        if not isinstance(key, str):
            raise TypeError, "Invalid entity data key: " + str(key)
        self.__edata[key] = value

    def getEntityData(self, key):
        """Retrieve the entity data value for a given key.

getEntityData(key):

Argument 'key' should be one of the keys returned from getEntityKeys().
        """
        return self.__edata[key]

    def setType(self, objtype):
        """Store the type of object in the dwgEntity instance.

setType(objtype)

Argument 'objtype' is an integer value corresponding to the entity
type. The OpenDWG specs give this information in more detail.
        """
        if not isinstance(objtype, int):
            raise TypeError, "Invalid object type: " + str(objtype)
        self.__cdata['TYPE'] = objtype

    def getType(self):
        """Return the type of object the dwgEntity represents.

getType()

This method returns an integer. See the OpenDWG specs for information
to match this value to the entity type.
        """
        return self.__cdata.get('TYPE')

    def setHandle(self, handle):
        """Set the handle (id) that the dwgEntity holds.

setHandle(handle)

Argument 'handle' is a tuple containing integer values.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['HANDLE'] = handle

    def getHandle(self):
        """Return the handle (id) of the dwgEntity.

getHandle()

This method returns a tuple containing integers.
        """
        return self.__cdata['HANDLE']

    def setVersion(self, version):
        """Set a version string in the dwgEntity.

setVersion(version)

Argument 'version' must be a string.
        """
        if not isinstance(version, str):
            raise TypeError, "Invalid version string: " + str(version)
        self.__cdata['VERSION'] =  version

    def getVersion(self):
        """Retrieve the version string in the dwgEntity.

getVersion()

This method returns a string, or None if the setVersion() method
has not been invoked on the instance.
        """
        return self.__cdata.get('VERSION')

    def setMode(self, mode):
        """Set the mode of the entity.

setMode(mode)

Argument 'mode' must be an integer.
        """
        if not isinstance(mode, int):
            raise TypeError, "Invalid mode: " + str(mode)
        self.__cdata['MODE'] = mode

    def getMode(self):
        """Return the mode of the entity.

getMode()

This method returns an integer value.
        """
        return self.__cdata.get('MODE')

    def setNumReactors(self, nr):
        """Set the number of reactors of an entity.

setNumReactors(nr)

Argument 'nr' must be an integer.
        """
        if not isinstance(nr, int):
            raise TypeError, "Invalind reactor count:" + str(nr)
        self.__cdata['NUMREACTORS'] = nr

    def getNumReactors(self):
        """Get the number of reactors of an entity.

getNumReactors()

This method returns an integer value.
        """
        return self.__cdata.get('NUMREACTORS', 0)

    def setNoLinks(self, flag):
        """Set the 'nolinks' flag of an entity.

setNoLinks(self, flag)

Argument 'flag' can be either True, False, or an integer. If
it is an integer, 0 is False, and all other values are True.
        """
        if isinstance(flag, int):
            if flag == 0:
                _nlflag = False
            else:
                _nlflag = True
        elif flag is False:
            _nlflag = False
        elif flag is True:
            _nlflag = True
        else:
            raise TypeError, "Invalid type for flag: " + str(flag)
        self.__cdata['NOLINKS'] = _nlflag

    def getNoLinks(self):
        """Return the 'nolinks' flag of an entity.

getNoLinks()
        """
        return self.__cdata.get('NOLINKS')

    def setIsLayerByLinetype(self, flag):
        """Set a flag value.

setIsLayerByLinetype(flag)

Argument 'flag' can be either True, False, or an integer. If
it is an integer, 0 is False, and all other values are True.
        """
        if isinstance(flag, int):
            if flag == 0:
                _iflag = False
            else:
                _iflag = True
        elif flag is False:
            _iflag = False
        elif flag is True:
            _iflag = True
        else:
            raise TypeError, "Invalid type for flag: " + str(flag)
        self.__cdata['ILBT'] = _iflag

    def getIsLayerByLinetype(self):
        """Return the flag value.

getIsLayerByLinetype()
        """
        return self.__cdata.get('ILBT')

    def setColor(self, color):
        """Store the entity color in the dwgEntity object.

setColor(color)

Argument 'color' is an integer.
        """
        if not isinstance(color, int):
            raise TypeError, "Invalid color: " + str(color)
        self.__cdata['COLOR'] = color

    def getColor(self):
        """Return the color of the entity.

getColor()

This method returns an integer giving the entity color.
        """
        return self.__cdata.get('COLOR')

    def setLinetypeScale(self, scale):
        """Store the linetype scale factor of the DWG object.

setLinetypeScale(scale)

Argument 'scale' must be a float value.
        """
        if not isinstance(scale, float):
            raise TypeError, "Invalid linetype scale: " + str(scale)
        self.__cdata['LTSCALE'] = scale

    def getLinetypeScale(self):
        """Return the linetype scale factor for the dwgEntity.

getLinetypeScale()

This method returns a float value.
        """
        return self.__cdata.get('LTSCALE')

    def setLinetypeFlags(self, flags):
        """Set the linetype flags.

setLinetypeFlags(flags)

Argument 'flags' must be an integer.
        """
        if not isinstance(flags, int):
            raise TypeError, "Invalid linetype flags: " + str(flags)
        self.__cdata['LTFLAGS'] = flags

    def getLinetypeFlags(self):
        """Return the linetype flags.

getLinetypesFlags()
        """
        return self.__cdata.get('LTFLAGS')

    def setPlotstyleFlags(self, flags):
        """Set the plotstyle flags.

setPlotstyleFlags(flags)

Argument 'flags' must be an integer.
        """
        if not isinstance(flags, int):
            raise TypeError, "Invalid plotstyle flags: " + str(flags)
        self.__cdata['PSFLAGS'] = flags

    def getPlotstyleFlags(self):
        """Get the plotstyle flags.

getPlotstyleFlags()
        """
        return self.__cdata.get('PSFLAGS')

    def setInvisiblity(self, flag):
        """Set the invisiblity flag.

setInvisiblity(flag)

Argument 'flag' can be either True, False, or an integer. If
it is an integer, 0 is False, and all other values are True.
        """
        if isinstance(flag, int):
            if flag == 0:
                _iflag = False
            else:
                _iflag = True
        elif flag is False:
            _iflag = False
        elif flag is True:
            _iflag = True
        else:
            raise TypeError, "Invalid type for flag: " + str(flag)
        self.__cdata['INVIS'] = _iflag

    def getInvisibility(self):
        """Get the invisibility flag.

getInvisibility()
        """
        return self.__cdata.get('INVIS')

    def setLineweight(self, weight):
        """Set the line weight.

setLineweight(weight)

Argument 'weight' must be an integer.
        """
        if not isinstance(weight, int):
            raise TypeError, "Invalid lineweight: " + str(weight)
        self.__cdata['LINEWEIGHT'] = weight

    def getLineweight(self):
        """Get the line weight.

getLineweight()
        """
        return self.__cdata.get('LINEWEIGHT')

    def setSubentity(self, handle): # code 3 handles
        """Set the subentity handle (id).

setSubentity(handle)

Argument 'handle' must be a tuple
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['SUBENTITY'] = handle

    def getSubentity(self):
        """Get the subentity handle.

getSubentity
        """
        return self.__cdata.get('SUBENTITY')

    def addReactor(self, handle): # code 4 handles
        """Add a reactor to an dwgEntity.

addReactor(handle)

Argument 'handle' must be a tuple.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        if 'REACTORS' not in self.__cdata:
            self.__cdata['REACTORS'] = []
        self.__cdata['REACTORS'].append(handle)

    def getReactors(self):
        """Get all the reactors for a dwgEntity.

getReactors()

This method returns a list of tuples.
        """
        _rlist = []
        if 'REACTORS' in self.__cdata:
            _rlist.extend(self.__cdata['REACTORS'][:])
        return _rlist

    def setXdicobj(self, handle): # code 3 handle
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['XDICOBJ'] = handle

    def getXdicobj(self):
        return self.__cdata.get('XDICOBJ')

    def setLayer(self, handle): # code 5 handle
        """Store the layer where the entity is held.

setLayer(handle)

Argument 'handle' is a tuple containing integers.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['LAYER'] = handle

    def getLayer(self):
        """Return the layer handle to which this entity belongs.

getLayer()

This method returns a tuple.
        """
        return self.__cdata.get('LAYER')

    def setLinetype(self, handle): # code 5 handle
        """Set the linetype handle for a dwgEntity.

setLinetype(handle)

Argument 'handle' must be a tuple.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['LINETYPE'] = handle

    def getLinetype(self):
        """Get the linetype for a dwgEntity.

getLinetype()
        """
        return self.__cdata.get('LINETYPE')

    def setPrevious(self, handle): # code 4 handle
        """Set the previous entity handle for a dwgEntity.

setPrevious(handle)

Argument 'handle' must be a tuple.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['PREV'] = handle

    def getPrevious(self):
        """Get the previous entity handle for a dwgEntity.

getPrevious()
        """
        return self.__cdata.get('PREV')

    def setNext(self, handle): # code 4 handle
        """Set the next entity handle for a dwgEntity.

setNext(handle)

Argument 'handle' must be a tuple.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['NEXT'] = handle

    def getNext(self):
        """Get the next entity handle for a dwgEntity.

getNext()
        """
        return self.__cdata.get('NEXT')

    def setPlotstyle(self, handle): # code 5 handle
        """Set the plotstyle handle for an dwgEntity.

setPlotstyle(handle)

Argument 'handle' must be a tuple.
        """
        if not isinstance(handle, tuple):
            raise TypeError, "Invalid handle: " + str(handle)
        self.__cdata['PLOTSTYLE'] = handle

    def getPlotstyle(self):
        """Get the plotstyle handle for a dwgEntity.

getPlotstyle()
        """
        return self.__cdata.get('PLOTSTYLE')



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# basic text functionality
#

import math
import types

import Generic.color
import Generic.util
import Generic.entity

#
# common constant for style and weight
#

NORMAL = 0

#
# styles
#

OBLIQUE = 1
ITALIC = 2

#
# weights
#

LIGHT = 1
BOLD = 2
HEAVY = 3

class TextFormat(Generic.entity.Entity):
    """A class for describing text string properties.

A TextFormat object has the following attributes:

family: The font family
style: The font style
size: The font size
weight: The font weight
color: The font color

A TextFormat object has the following methods:

{get/set}Family(): Get/Set the font family.
{get/set}Size(): Get/Set the font size.
{get/set}Style(): Get/Set the font style.
{get/set}Weight(): Get/Set the font weight.
{get/set}Color(): Get/Set the font color.
    """
    messages = {
        'fontfamily_changed' : True,
        'fontsize_changed' : True,
        'fontstyle_changed' : True,
        'fontweight_changed' : True,
        'fontcolor_changed' : True,
        }
    #
    # default text color
    #

    __defcolor = Generic.color.Color(0xffffff)

    def __init__(self, family='Sans', size=12, style=NORMAL, weight=NORMAL,
                 color=None):
        """Initialize a TextFormat object.

ts = TextFormat([family, size, style, weight, color])

Each argument has a default value:

family: Sans
size: 12
style: NORMAL
weight: NORMAL
color: A default color - #ffffff
        """
        _family = family
        if not isinstance(_family, types.StringTypes):
            raise TypeError, "Invalid font family name: " + str(_family)
        _size = size
        if not isinstance(_size, int):
            _size = int(size)
        if _size < 1:
            raise ValueError, "Invalid text size: %d" % _size
        _style = style
        if (_style != NORMAL and
            _style != OBLIQUE and
            _style != ITALIC):
            raise ValueError, "Invalid text style: " + str(style)
        _weight = weight
        if (_weight != NORMAL and
            _weight != LIGHT and
            _weight != BOLD and
            _weight != HEAVY):
            raise ValueError, "Invalid text weight: " + str(weight)
        _color = color
        if _color is None:
            _color = TextFormat.__defcolor
        if not isinstance(color, Generic.color.Color):
            _color = Generic.color.Color(color)
        Generic.entity.Entity.__init__(self)
        self.__family = _family
        self.__size = _size
        self.__fstyle = _style
        self.__weight = _weight
        self.__color = _color

    def __eq__(self, tf):
        """Compare to TextFormats for equality.
        """
        if not isinstance(tf, TextFormat):
            raise TypeError, "Invalid TextFormat: " + str(tf)
        _val = False
        if (self.__family == tf.getFamily() and
            self.__size == tf.getSize() and
            self.__fstyle == tf.getStyle() and
            self.__weight == tf.getWeight() and
            self.__color == tf.getColor()):
            _val = True
        return _val

    def __ne__(self, tf):
        """Compare to TextFormats for inequality.
        """
        if not isinstance(tf, TextFormat):
            raise TypeError, "Invalid TextFormat: " + str(tf)
        _val = True
        if (self.__family == tf.getFamily() and
            self.__size == tf.getSize() and
            self.__fstyle == tf.getStyle() and
            self.__weight == tf.getWeight() and
            self.__color == tf.getColor()):
            _val = False
        return _val

    def getFamily(self):
        """Return the font family.

getFamily()
        """
        return self.__family

    def setFamily(self, family): # what should default be?
        """Set the font family.

setFamily(family)
        """
        if self.isLocked():
            raise RuntimeError, "Setting family not allowed - object locked."
        if not isinstance(family, str):
            raise TypeError, "Invalid font family name: " + str(family)
        _f = self.__family
        if _f != family:
            self.__family = family
            self.sendMessage('fontfamily_changed', _f)
            self.modified()

    family = property(getFamily, setFamily, None, "Text font family")

    def getSize(self):
        """Return the font size.

getSize()
        """
        return self.__size

    def setSize(self, size=12):
        """Set the font size.

setSize([size])

The argument "size" should be an integer value
greater than 0. If this method is called without
arguments, the font size defaults to 12.
        """
        if self.isLocked():
            raise RuntimeError, "Setting size not allowed - object locked."
        _size = size
        if not isinstance(_size, int):
            _size = int(size)
        if _size < 1:
            raise ValueError, "Invalid text size: %d" % _size
        _s = self.__size
        if _s != _size:
            self.__size = _size
            self.sendMessage('fontsize_changed', _s)
            self.modified()

    size = property(getSize, setSize, None, "Text font size")

    def getStyle(self):
        """Return the font style.

getStyle()
        """
        return self.__fstyle

    def setStyle(self, style=NORMAL):
        """Set the font style.

setStyle([style])

The argument "style" should be one of the following:

NORMAL
OBLIQUE
ITALIC

If the method is invoked without an argument, the font
style is set to NORMAL.
        """
        if self.isLocked():
            raise RuntimeError, "Setting style not allowed - object locked."
        _style = style
        if (_style != NORMAL and
            _style != OBLIQUE and
            _style != ITALIC):
            raise ValueError, "Invalid text style: " + str(style)
        _s = self.__fstyle
        if _s != _style:
            self.__fstyle = _style
            self.sendMessage('fontstyle_changed', _s)
            self.modified()

    style = property(getStyle, setStyle, None, "Text font style")

    def getWeight(self):
        """Return the font weight.

getWeight()
        """
        return self.__weight

    def setWeight(self, weight=NORMAL):
        """Set the font weight.

setWeight([weight])

The argument "weight" should be one of the following:

NORMAL
LIGHT
BOLD
HEAVY

If the method is called without an argument, the weight
is set to NORMAL.
        """
        if self.isLocked():
            raise RuntimeError, "Setting weight not allowed - object locked."
        _weight = weight
        if (_weight != NORMAL and
            _weight != LIGHT and
            _weight != BOLD and
            _weight != HEAVY):
            raise ValueError, "Invalid text weight: " + str(weight)
        _w = self.__weight
        if _w != _weight:
            self.__weight = _weight
            self.sendMessage('fontweight_changed', _w)
            self.modified()

    weight = property(getWeight, setWeight, None, "Text font weight")

    def getColor(self):
        """Return the font color.

getColor()
        """
        return self.__color

    def setColor(self, color=None):
        """Set the font color.

setColor([color])

The argument "color" should be a Color object. If the method
is called without an argument, the font color is set to the
default text color.
        """
        if self.isLocked():
            raise RuntimeError, "Setting color not allowed - object locked."
        _color = color
        if _color is None:
            _color = TextFormat.__color
        if not isinstance(_color, Generic.color.Color):
            _color = Generic.color.Color(color)
        _c = self.__color
        if _c != _color:
            self.__color = _color
            self.sendMessage('fontcolor_changed', _c)
            self.modified()

    color = property(getColor, setColor, None, "Text color")

    def sendsMessage(self, m):
        if m in TextFormat.messages:
            return True
        return Generic.entity.Entity.sendsMessage(self, m)

def font_style_string(style):
    """Return a text string for the font style.

font_style_string(style)
    """
    if style == NORMAL:
        _str = 'normal'
    elif style == OBLIQUE:
        _str = 'oblique'
    elif style == ITALIC:
        _str = 'italic'
    else:
        raise ValueError, "Unknown font style: " + str(style)
    return _str

def font_weight_string(weight):
    """Return a text string for the font weight.

font_weight_string(weight)
    """
    if weight == NORMAL:
        _str = 'normal'
    elif weight == LIGHT:
        _str = 'light'
    elif weight == BOLD:
        _str = 'bold'
    elif weight == HEAVY:
        _str = 'heavy'
    else:
        raise ValueError, "Unknown font weight: " + str(weight)
    return _str

#
# line justification options
#

ALIGN_LEFT = 0
ALIGN_CENTER = 1
ALIGN_RIGHT = 2

class TextStyle(TextFormat):
    """A class for storing named TextFormat objects.

The TextStyle class is an extension of the TextFormat class
with the a 'name' used for referencing the properties stored
in an instance of this class. As this class is derived from
the TextFormat class, it shares all the attributes and methods
of that class, except once the values in the object are set
they cannot be changed.

The TextStyle class has the following extra attributes:

name: The name of the TextStyle object.

The TextStyle class has th following extra methods:

{get/set}Name(): Get/Set the name of the TextStyle.
{get/set}Alignment(): Get/Set the line justification.
    """

    #
    # default text color
    #

    __defcolor = Generic.color.Color(0xffffff)

    def __init__(self, name, family="Sans", size=12,
                 style=NORMAL, weight=NORMAL, color=None):
        if not isinstance(name, types.StringTypes):
            raise ValueError, "Invalid name: " + str(name)
        self.__name = None
        _name = name
        if isinstance(_name, str):
            _name = unicode(name)
        _color = color
        if _color is None:
            _color = TextStyle.__defcolor
        self.__initialized = False
        TextFormat.__init__(self, family, size, style, weight, _color)
        self.__name = _name
        self.__alignment = ALIGN_LEFT
        self._finalize()

    def _finalize(self):
        self.__initialized = True

    def __eq__(self, ts):
        """Test two TextStyles for equality.
        """
        if not isinstance(ts, TextStyle):
            raise TypeError, "Invalid TextStyle: " + str(ts)
        _val = False
        if (self.__name == ts.getName() and
            self.__alignment == ts.getAlignment() and
            TextFormat.__eq__(self, ts)):
            _val = True
        return _val

    def __ne__(self, ts):
        """Test two TextStyles for inequality.
        """
        if not isinstance(ts, TextStyle):
            raise TypeError, "Invalid TextStyle: " + str(ts)
        _val = True
        if (self.__name == ts.getName() and
            self.__alignment == ts.getAlignment() and
            TextFormat.__eq__(self, ts)):
            _val = False
        return _val
#
# override TextFormat setXXXX() methods so the values in the
# TextStyle cannot be changed (easily)
#
    def setFamily(self, family):
        if not self.__initialized:
            TextFormat.setFamily(self, family)

    family = property(TextFormat.getFamily, setFamily, None, "Style family.")

    def setColor(self, color):
        if not self.__initialized:
            TextFormat.setColor(self, color)

    color = property(TextFormat.getColor, setColor, None, "Style color.")

    def setWeight(self, weight):
        if not self.__initialized:
            TextFormat.setWeight(self, weight)

    weight = property(TextFormat.getWeight, setWeight, None, "Style weight.")

    def setStyle(self, style):
        if not self.__initialized:
            TextFormat.setStyle(self, style)

    weight = property(TextFormat.getStyle, setStyle, None, "Style font style.")

    def setSize(self, size):
        if not self.__initialized:
            TextFormat.setSize(self, size)

    size = property(TextFormat.getSize, setSize, None, "Style font size.")

    def getName(self):
        """Retrieve the name of the TextStyle.

getName()
        """
        return self.__name

    def setName(self, name):
        """Set the name of the TextStyle.

setName(name)

The argument "name" should be a string or unicode string.
        """
        if not isinstance(name, types.StringTypes):
            raise ValueError, "Invalid name: " + str(name)
        _name = name
        if isinstance(_name, str):
            _name = unicode(name)
        self.__name = _name

    name = property(getName, setName, None, "TextStyle name.")


    def setAlignment(self, opt):
        """Set left, center, or right line justification.

setAlignment(opt)

The argument "opt" should be one of

Text.ALIGN_LEFT
Text.ALIGN_CENTER
Text.ALIGN_RIGHT
        """
        if (opt != ALIGN_LEFT and
            opt != ALIGN_CENTER and
            opt != ALIGN_RIGHT):
            raise ValueError, "Invalid justification: " + str(opt)
        self.__alignment = opt

    def getAlignment(self):
        """Return the line justification setting.

getAlignment()
        """
        return self.__alignment
#
# the font_prop_map and parse_font() should be moved as
# they are GTK specific ...
#

font_prop_map = {
    'Oblique' : 'style',
    'Italic' : 'style',
    'Ultra-Light' : 'weight',
    'Light' : 'weight',
    'Medium' : 'weight',
    'Semi-Bold' : 'weight',
    'Bold' : 'weight',
    'Ultra-Bold' : 'weight',
    'Heavy' : 'weight',
    'Ultra-Condensed' : 'stretch',
    'Extra-Condensed' : 'stretch',
    'Condensed' : 'stretch',
    'Semi-Condensed' : 'stretch',
    'Semi-Expanded' : 'stretch',
    'Expanded' : 'stretch',
    'Extra-Expanded' : 'stretch',
    'Ultra-Expanded' : 'stretch',
    }

def parse_font(fontstr):
    _size = 12
    _weight = NORMAL
    _style = NORMAL
    _stretch = NORMAL
    _family = 'Sans'
    if fontstr != '':
        _fontlist = fontstr.split()
        _fontlist.reverse()
        if _fontlist[0].isdigit():
            _sz = _fontlist.pop(0)
            _size = int(_sz)
        while (_fontlist[0] in font_prop_map):
            _prop = _fontlist.pop(0)
            _item = font_prop_map[_prop]
            # print "prop: " + _prop
            # print "item: " + _item
            if _item == 'style':
                if _prop == 'Oblique':
                    _style = 1
                elif _prop == 'Italic':
                    _style = 2
                else:
                    _style = NORMAL # default
            elif _item == 'weight':
                if (_prop == 'Ultra-Light' or
                    _prop == 'Light' or
                    _prop == 'Medium'):
                    _weight = 1
                elif (_prop == 'Semi-Bold' or
                      _prop == 'Bold'):
                    _weight = 2
                elif (_prop == 'Ultra-Bold' or
                      _prop == 'Heavy'):
                    _weight = 3
                else:
                    _weight = NORMAL
            elif _item == 'stretch':
                _stretch = _prop # fixme - add stretching bits
            else:
                raise ValueError, "Unknown font property: " + _item
        _fontlist.reverse()
        if len(_fontlist):
            _family = ' '.join(_fontlist)
    return (_family, _style, _weight, _stretch, _size)

#
# The TextBlock class defines the routines used to store unicode text
# objects in an image.
#

class TextBlock(TextFormat):
    """A class for storing text in an image

The TextBlock class is derived from the TextFormat class, so it
shares the methods and attributes of that class. In addtion
to the methods in the base class, a TextBlock has the following
addtional methods:

{get/set}Style(): Get/Set the TextStyle used by the TextBlock.
applyStyle(): Set the TextBlock properties to those in the style.
{get/set}Location(): Get/Set the Location of the TextBlock.
{get/set}Angle(): Get/Set the inclination angle of the TextBlock.
    """

    messages = {
    'style_changed' : True,
    'angle_changed' : True,
    'text_changed' : True,
    'moved' : True
    }
    
    def __init__(self, text, style):
        TextFormat.__init__(self) # just get defaults ...
        _text = text
        if not isinstance(_text, unicode):
            _text = unicode(text)
        if not isinstance(style, TextStyle):
            raise TypeError, "Invalid TextStyle: " + str(style)
        self.__text = _text
        self.__location = None
        self.__angle = 0.0
        self.__textstyle = style

    def __eq__(self, tb):
        if not isinstance(tb, TextBlock):
            raise TypeError, "Invalid TextBlock: " + str(tb)
        _stext = self.__text
        _sl = self.__location
        _tbtext = tb.getText()
        _tbl = tb.getLocation()
        _val = False
        if _stext == _tbtext:
            if _sl is None and _tbl is None:
                _val = True
            elif _sl is not None and _tbl is not None:
                _sx, _sy = _sl
                _tx, _ty = _tbl
                if abs(_sx - _tx) < 1e-10 and abs(_sy - _ty) < 1e-10:
                    _val = True
        return _val

    def __ne__(self, tb):
        if not isinstance(tb, TextBlock):
            raise TypeError, "Invalid TextBlock: " + str(tb)
        _stext = self.__text
        _sl = self.__location
        _tbtext = tb.getText()
        _tbl = tb.getLocation()
        _val = True
        if _stext == _tbtext:
            if _sl is None and _tbl is None:
                _val = False
            elif _sl is not None and _tbl is not None:
                _sx, _sy = _sl
                _tx, _ty = _tbl
                if abs(_sx - _tx) < 1e-10 and abs(_sy - _ty) < 1e-10:
                    _val = False
        return _val

    def getText(self):
        """Retrieve the text stored in the TextBlock.

getText()
        """
        return self.__text

    def setText(self, text):
        """Set the text stored in the TextBlock.

setText(text)

The argument 'text' is converted to a unicode object if
it is not one already.
        """
        _text = text
        if not isinstance(_text, unicode):
            _text = unicode(text)
        _t = self.__text
        if _t != _text:
            self.__text = _text
            self.sendMessage('text_changed', _t)
            self.modified()

    def getTextStyle(self):
        """Get the style used by the TextBlock.

getTextStyle()

This method returns a TextStyle object.
        """
        return self.__textstyle

    def setTextStyle(self, style):
        """Set the style used by the TextBlock.

setTextStyle(style)

The argument "style" must be a TextStyle object.
        """
        if not isinstance(style, TextStyle):
            raise TypeError, "Invalid TextStyle: " + str(style)
        _s = self.__textstyle
        if _s != style:
            self.__textstyle = style
            self.applyStyle()
            self.sendMessage('style_changed', _s)
            self.modified()

    def applyStyle(self):
        """Set the TextBlock properties to those in the TextStyle.

applyStyle()

Invoking this method will set all the TextBlock properties to those
defined in the TextStyle.
        """
        if self.__textstyle is None:
            raise ValueError, "Undefined style"
        _style = self.__textstyle
        self.setFamily(_style.getFamily())
        self.setWeight(_style.getWeight())
        self.setStyle(_style.getStyle())
        self.setSize(_style.getSize())
        self.setColor(_style.getColor())

    def getLocation(self):
        """Get the location where the text block is located.

getLocation()

This method returns a tuple giving the x/y coordinates of the
block location, or None if no location has been set.
        """
        return self.__location

    def setLocation(self, x, y):
        """Set the location where the text block should be placed.

setLocation(x, y)

The arguments x and y should both be floats.
        """
        _x, _y = Generic.util.make_coords(x, y)
        _loc = self.__location
        _sx = _x + 1.0 # this guarantees the location will be set
        _sy = _y + 1.0 # so does this ...
        if _loc is not None:
            _sx, _sy = _loc
        if abs(_x - _sx) > 1e-10 or abs(_y - _sy) > 1e-10:
            self.__location = (_x, _y)
            self.sendMessage('moved', _x, _y)
            self.modified()

    def getAngle(self):
        """Return the angle at which the text block is drawn.

getAngle()

This method returns an angle -360.0 < angle < 360.0.
        """
        return self.__angle

    def setAngle(self, angle):
        """Set the angle at which the text block should be drawn.

setAngle(angle)

        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        if _angle > 360.0 or _angle < -360.0:
            _angle = math.fmod(value, 360.0)
        _a = self.__angle
        if abs(_a - _angle) > 1e-10:
            self.__angle = _angle
            self.sendMessage('angle_changed', _a)
            self.modified()

    def sendsMessage(self, m):
        if m in TextBlock.messages:
            return True
        return TextFormat.sendsMessage(self, m)



#
# Copyright (c) 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for leader lines
#

from __future__ import generators

import math
import array

import graphicobject
import entity
import tolerance
import style
import linetype
import color
import point
import util
import quadtree

class Leader(graphicobject.GraphicObject):
    """A class representing a leader line.

A leader line is usually used as a visual connector between
some text and an entity in a drawing. Leader lines cannot be
used to define an edge of some shape.

A Leader object has the following attributes:
p1: A Point object representing the first end point.
p2: A Point object representing the leader mid point.
p3: A Point object representing the final end point.
arrowsize: The size of the arrow at the end of the leader line

A Leader object has the following methods:

getPoints(): Return the points defining the Leader
{get/set}P1(): Get/Set the Leader first endpoint.
{get/set}P2(): Get/Set the Leader midpoint
{get/set}P3(): Get/Set the Leader final endpoint
{get/set}ArrowSize(): Get/Set the Leader line arrow size.
calcArrowPoints(): Calculate where the Leader arrow points are.
getArrowPoints(): Return where the Leader arrow points are.
move(): Move the Leader.
mapCoords(): See if a point lies within some distance of a Leader.
inRegion(): Test if the Leader is visible in some area.
clone(): Make an identical copy of a Leader.
    """
    __defstyle = style.Style(u'Leader Default Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)
    messages = {
    'moved' : True,
    'point_changed' : True,
    'size_changed' : True,
    }

    def __init__(self, p1, p2, p3, size=1.0,
                 st=None, lt=None, col=None, th=None, **kw):
        """Initialize a Leader object.

Leader(p1, p2, p3[, size, st, lt, col, th])

The following arguments are required:

p1: Leader first endpoint - may be a Point or a two-item tuple of floats
p2: Leader mid point - may be a Point or a two-item tuple of floats
p3: Leader final endpoint - may be a Point or a two-item tuple of floats

Argument size is optional. It gives the size of the arrow at
the end of the leader line, and defaults to 1.0.
        """
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 is _p2:
            raise ValueError, "Leader points p1 and p2 cannot be identical."
        _p3 = p3
        if not isinstance(_p3, point.Point):
            _p3 = point.Point(_p3)
        if _p1 is _p3:
            raise ValueError, "Leader points p1 and p3 cannot be identical."
        if _p2 is _p3:
            raise ValueError, "Leader points p2 and p3 cannot be identical"
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        _st = st
        if _st is None:
            _st = Leader.__defstyle
        graphicobject.GraphicObject.__init__(self, _st, lt, col, th, **kw)
        self.__p1 = _p1
        _p1.connect('moved', self, Leader.movePoint)
        _p1.storeUser(self)
        self.__p2 = _p2
        _p2.connect('moved', self, Leader.movePoint)
        _p2.storeUser(self)
        self.__p3 = _p3
        _p3.connect('moved', self, Leader.movePoint)
        _p3.storeUser(self)
        self.__size = _size
        self.__arrow_pts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        self.calcArrowPoints()

    def __str__(self):
        return "Leader: %s to %s to %s" % (self.__p1, self.__p2, self.__p3)

    def __eq__(self, obj):
        """Compare two leader lines for equality.
        """
        if not isinstance(obj, Leader):
            return False
        if obj is self:
            return True
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getPoints()
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            return True
        return False

    def __ne__(self, obj):
        """Compare two leader lines for inequality.
        """
        if not isinstance(obj, Leader):
            return True
        if obj is self:
            return False
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getPoints()
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            return False
        return True

    def getValues(self):
        _pids = entity.Entity.getValues(self)
        _gv = self.getGraphicValues(Leader.__defstyle)
        return ('leader', _pids, _gv, self.__p1.getID(), self.__p2.getID(),
                self.__p3.getID(), self.__size)

    def getPoints(self):
        """Get the points defining the Leader.

getPoints()

This function returns a tuple containing the three Point objects
that are the defining points of the Leader
        """
        return self.__p1, self.__p2, self.__p3

    def getP1(self):
        """Return the first endpoint Point of the Leader

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint Point of the Leader

setP1(p)

The argument p should be a Point.
        """
        if self.isLocked():
            raise RuntimeError, "Setting points not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid P1 point: " + str(p)
        if p is self.__p2 or p is self.__p3:
            raise ValueError, "Leader points cannot be identical."
        _pt = self.__p1
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p1 = p
            self.sendMessage('point_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Leader.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                _x2, _y2 = self.__p2.getCoords()
                _x3, _y3 = self.__p3.getCoords()
                self.sendMessage('moved', _pt.x, _pt.y, _x2, _y2, _x3, _y3)
            self.modified()

    p1 = property(getP1, setP1, None, "First endpoint of the Leader.")

    def getP2(self):
        """Return the midpoint Point of the Leader.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the midpoint Point of the Leader.

setP2(p)

The argument p should be a Point.
        """
        if self.isLocked():
            raise RuntimeError, "Setting points not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid P1 point: " + str(p)
        if p is self.__p1 or p is self.__p3:
            raise ValueError, "Leader points cannot be identical."
        _pt = self.__p2
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p2 = p
            self.sendMessage('point_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Leader.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                self.calcArrowPoints()
                _x1, _y1 = self.__p1.getCoords()
                _x3, _y3 = self.__p3.getCoords()
                self.sendMessage('moved', _x1, _y1, _pt.x, _pt.y, _x3, _y3)
            self.modified()

    p2 = property(getP2, setP2, None, "Leader midpoint.")

    def getP3(self):
        """Return the final Point of the Leader.

getP3()
        """
        return self.__p3

    def setP3(self, p):
        """Set the final endpoint Point of the Leader.

setP3(p)

The argument "p" should be a Point.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid Point for p3 endpoint: " + `p`
        if self.isLocked():
            raise RuntimeError, "Setting points not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid P1 point: " + str(p)
        if p is self.__p1 or p is self.__p2:
            raise ValueError, "Leader points cannot be identical."
        _pt = self.__p3
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p3 = p
            self.sendMessage('point_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Leader.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                self.calcArrowPoints()
                _x1, _y1 = self.__p1.getCoords()
                _x2, _y2 = self.__p2.getCoords()
                self.sendMessage('moved', _x1, _y1, _x2, _y2, _pt.x, _pt.y)
            self.modified()

    p3 = property(getP3, setP3, None, "Third endpoint of the Leader.")

    def getArrowSize(self):
        """Return the size of the leader line arrow.

getArrowSize()
        """
        return self.__size

    def setArrowSize(self, size):
        """Set the size of the leader line arrow.

setSize(size)

Argument "size" should be a float greater than or equal to 0.0.
        """
        if self.isLocked():
            raise RuntimeError, "Cannot change arrow size - object locked."
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        _os = self.__size
        if abs(_os - _size) > 1e-10:
            self.__size = _size
            self.calcArrowPoints()
            self.sendMessage('size_changed', _os)
            self.modified()

    arrowsize = property(getArrowSize, setArrowSize,
                    None, "Leader line arrow size.")

    def calcArrowPoints(self):
        """Calculate where the Leader arrow points are.

calcArrowPoints()
        """
        _x1, _y1 = self.__p2.getCoords()
        _x2, _y2 = self.__p3.getCoords()
        if abs(_x2 - _x1) < 1e-10: # vertical
            _cosine = 0.0
            if _y2 > _y1:
                _sine = 1.0
            else:
                _sine = -1.0

        elif abs(_y2 - _y1) < 1e-10: # horizontal
            _sine = 0.0
            if _x2 > _x1:
                _cosine = 1.0
            else:
                _cosine = -1.0
        else:
            _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
            _sine = math.sin(_angle)
            _cosine = math.cos(_angle)
        _size = self.__size
        _height = _size/5.0
        # p1 -> (x,y) = (-size, _height)
        self.__arrow_pts[0] = (_cosine * (-_size) - _sine * _height) + _x2
        self.__arrow_pts[1] = (_sine * (-_size) + _cosine * _height) + _y2
        # p2 -> (x,y) = (-size, -_height)
        self.__arrow_pts[2] = (_cosine * (-_size) - _sine *(-_height)) + _x2
        self.__arrow_pts[3] = (_sine * (-_size) + _cosine *(-_height)) + _y2

    def getArrowPoints(self):
        """Return the endpoints of the Leader arrow.

getArrowPoints()

This method returns an array holding four float values.
        """
        return self.__arrow_pts

    def move(self, dx, dy):
        """Move a Leader.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if (self.isLocked() or
            self.__p1.isLocked() or
            self.__p2.isLocked() or
            self.__p3.isLocked()):
            raise RuntimeError, "Moving Leader not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x1, _y1 = self.__p1.getCoords()
            _x2, _y2 = self.__p2.getCoords()
            _x3, _y3 = self.__p3.getCoords()
            self.ignore('moved')
            self.__p1.move(_dx, _dy)
            self.__p2.move(_dx, _dy)
            self.__p3.move(_dx, _dy)
            self.receive('moved')
            self.calcArrowPoints()
            self.sendMessage('moved', _x1, _y1, _x2, _y2, _x3, _y3)
            self.modified()

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Leader to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the Leader. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        _pt = util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)
        if _pt is None:
            _pt = util.map_coords(_x, _y, _x2, _y2, _x3, _y3, _t)
        if _pt is not None:
            return _pt
        return None

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a Leader exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the Leader lies within that area. If
the optional argument fully is used and is True, then all
the Leader points must lie within the boundary. Otherwise,
the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        _pxmin = min(_x1, _x2, _x3)
        _pymin = min(_y1, _y2, _y3)
        _pxmax = max(_x1, _x2, _x3)
        _pymax = max(_y1, _y2, _y3)
        if ((_pxmax < _xmin) or
            (_pymax < _ymin) or
            (_pxmin > _xmax) or
            (_pymin > _ymax)):
            return False
        if fully:
            if ((_pxmin > _xmin) and
                (_pymin > _ymin) and
                (_pxmax < _xmax) and
                (_pymax < _ymax)):
                return True
            return False
        if util.in_region(_x1, _y1, _x2, _y2, _xmin, _ymin, _xmax, _ymax):
            return True
        return util.in_region(_x2, _y2, _x3, _y3, _xmin, _ymin, _xmax, _ymax)

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _p1 = self.__p1
        _p2 = self.__p2
        _p3 = self.__p3
        if p is _p1:
            _x1 = _x
            _y1 = _y
            _x2, _y2 = _p2.getCoords()
            _x3, _y3 = _p3.getCoords()
        elif p is _p2:
            _x1, _y1 = _p1.getCoords()
            _x2 = _x
            _y2 = _y
            _x3, _y3 = _p3.getCoords()
        elif p is _p3:
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            _x3 = _x
            _y3 = _y
        else:
            raise ValueError, "Unexpected Segment endpoint: " + `p`
        self.sendMessage('moved', _x1, _y1, _x2, _y2, _x3, _y3)
        self.modified()

    def clone(self):
        """Create an identical copy of a Leader.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        _cp3 = self.__p3.clone()
        _size = self.__size
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Leader(_cp1, _cp2, _cp3, _size, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Leader.messages:
            return True
        return graphicobject.GraphicObject.sendsMessage(self, m)

#
# Quadtree Leader storage
#

class LeaderQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 4:
            raise ValueError, "Expected 4 arguments, got %d" % _alen
        _lxmin, _lymin, _lxmax, _lymax = args        
        if not isinstance(_lxmin, float):
            _lxmin = float(args[0])
        if not isinstance(_lymin, float):
            _lymin = float(args[1])
        if not isinstance(_lxmax, float):
            _lxmax = float(args[2])
        if not isinstance(_lymax, float):
            _lymax = float(args[3])
        if not _lxmax > _lxmin:
            raise ValueError, "xmax not greater than xmin"
        if not _lymax > _lymin:
            raise ValueError, "ymax not greater than ymin"
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_lxmin > _xmax) or
                (_lxmax < _xmin) or
                (_lymin > _ymax) or
                (_lymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _lxmax < _xmid: # leader on left side
                    _ne = _se = False
                if _lxmin > _xmid: # leader on right side
                    _nw = _sw = False
                if _lymax < _ymid: # leader below
                    _nw = _ne = False
                if _lymin > _ymid: # leader above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, Leader):
            raise TypeError, "Invalid Leader object: " + `obj`
        if obj in self:
            return
        _p1, _p2, _p3 = obj.getPoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _x3, _y3 = _p3.getCoords()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _lxmin - 1.0
            _ymin = _lymin - 1.0
            _xmax = _lxmax + 1.0
            _ymax = _lymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _lxmin < _xmin:
                _xmin = _lxmin - 1.0
                _resize = True
            if _lxmax > _xmax:
                _xmax = _lxmax + 1.0
                _resize = True
            if _lymin < _ymin:
                _ymin = _lymin - 1.0
                _resize = True
            if _lymax > _ymax:
                _ymax = _lymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_lxmin, _lymin, _lxmax, _lymax):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, LeaderQuadtree.moveLeader)

    def old_addObject(self, obj):
        if not isinstance(obj, Leader):
            raise TypeError, "Invalid Leader object: " + `obj`
        if obj in self:
            return
        _p1, _p2, _p3 = obj.getPoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _x3, _y3 = _p3.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _lxmin - 1.0
            _ymin = _lymin - 1.0
            _xmax = _lxmax + 1.0
            _ymax = _lymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _lxmin < _xmin:
                _xmin = _lxmin - 1.0
                _resize = True
            if _lxmax > _xmax:
                _xmax = _lxmax + 1.0
                _resize = True
            if _lymin < _ymin:
                _ymin = _lymin - 1.0
                _resize = True
            if _lymax > _ymax:
                _ymax = _lymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_lxmin > _xmax) or
                (_lxmax < _xmin) or
                (_lymin > _ymax) or
                (_lymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _lxmax < _xmid: # leader on left side
                    _ne = _se = False
                if _lxmin > _xmid: # leader on right side
                    _nw = _sw = False
                if _lymax < _ymid: # leader below
                    _nw = _ne = False
                if _lymin > _ymid: # leader above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, LeaderQuadtree.moveLeader)

    def delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2, _p3 = obj.getPoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _x3, _y3 = _p3.getCoords()
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _pdict = {}
        for _node in self.getNodes(_lxmin, _lymin, _lxmax, _lymax):
            _node.delObject(obj) # leader may not be in the node ...
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2, _p3 = obj.getPoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _x3, _y3 = _p3.getCoords()
        _xmin = _ymin = _xmax = _ymax = None
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_lxmin > _xmax) or
                (_lxmax < _xmin) or
                (_lymin > _ymax) or
                (_lymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _lxmax < _xmid: # leader on left side
                    _ne = _se = False
                if _lxmin > _xmid: # leader on right side
                    _nw = _sw = False
                if _lymax < _ymid: # leader below
                    _nw = _ne = False
                if _lymin > _ymid: # leader above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # leader may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 6:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _x3 = args[4]
        if not isinstance(_x3, float):
            _x3 = float(args[4])
        _y3 = args[5]
        if not isinstance(_y3, float):
            _y3 = float(args[5])
        _t = tolerance.TOL
        if _alen > 6:
            _t = tolerance.toltest(args[6])
        if not len(self):
            return None
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _ldict = {}
        _bailout = False
        _leader = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_lxmin > _xmax) or
                (_lxmax < _xmin) or
                (_lymin > _ymax) or
                (_lymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _lxmax < (_xmid - _t): # leader on left side
                    _ne = _se = False
                if _lxmin > (_xmid + _t): # leader on right side
                    _nw = _sw = False
                if _lymax < (_ymid - _t): # leader below
                    _nw = _ne = False
                if _lymin > (_ymid + _t): # leader above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _l in _node.getObjects():
                    _lid = id(_l)
                    if _lid not in _ldict:
                        _p1, _p2, _p3 = _l.getPoints()
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x1) < _t) and (abs(_py - _y1) < _t)):
                            _px, _py = _p2.getCoords()
                            if ((abs(_px - _x2) < _t) and
                                (abs(_py - _y2) < _t)):
                                _px, _py = _p3.getCoords()
                                if ((abs(_px - _x2) < _t) and
                                    (abs(_py - _y2) < _t)):
                                    _bailout = True
                                    _leader = _l
                                    break
                        _ldict[_lid] = True
            if _bailout:
                break
        return _leader

    def moveLeader(self, obj, *args):
        if obj not in self:
            raise ValueError, "Leader not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 6:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _x3 = args[4]
        if not isinstance(_x3, float):
            _x3 = float(args[4])
        _y3 = args[5]
        if not isinstance(_y3, float):
            _y3 = float(args[5])
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        for _node in self.getNodes(_lxmin, _lymin, _lxmax, _lymax):
            _node.delObject(obj) # leader may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveLeader(self, obj, *args):
        if obj not in self:
            raise ValueError, "Leader not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 6:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _x3 = args[4]
        if not isinstance(_x3, float):
            _x3 = float(args[4])
        _y3 = args[5]
        if not isinstance(_y3, float):
            _y3 = float(args[5])
        _lxmin = min(_x1, _x2, _x3)
        _lxmax = max(_x1, _x2, _x3)
        _lymin = min(_y1, _y2, _y3)
        _lymax = max(_y1, _y2, _y3)
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_lxmin > _xmax) or
                (_lxmax < _xmin) or
                (_lymin > _ymax) or
                (_lymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _lxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _lxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _lymax < _ymid: # seg below
                    _nw = _ne = False
                if _lymin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # leader may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _leader = _tsep = None
        _bailout = False
        _ldict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _l in _node.getObjects():
                    _lid = id(_l)
                    _p1, _p2, _p3 = _l.getPoints()
                    if _lid not in _ldict:
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _leader = _l
                            _bailout = True
                            break
                        _px, _py = _p2.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _leader = _l
                            _bailout = True
                            break
                        _px, _py = _p3.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _leader = _l
                            _bailout = True
                            break
                        _ldict[_lid] = True
                    _pt = _l.mapCoords(_x, _y, _t)
                    if _pt is not None:
                        _px, _py = _pt
                        _sep = math.hypot((_px - _x), (_py - _y))
                        if _tsep is None:
                            _tsep = _sep
                            _leader = _l
                        else:
                            if _sep < _tsep:
                                _tsep = _sep
                                _leader = _l
            if _bailout:
                break
        return _leader

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _leaders = []
        if not len(self):
            return _leaders
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _ldict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _lxmin, _lymin, _lxmax, _lymax = _subnode.getBoundary()
                    if ((_lxmin > _xmax) or
                        (_lymin > _ymax) or
                        (_lxmax < _xmin) or
                        (_lymax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _l in _node.getObjects():
                    _lid = id(_l)
                    if _lid not in _ldict:
                        if _l.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _leaders.append(_l)
                        _ldict[_lid] = True
        return _leaders

#
# Leader history class
#

class LeaderLog(graphicobject.GraphicObjectLog):
    def __init__(self, l):
        if not isinstance(l, Leader):
            raise TypeError, "Invalid leader: " + `l`
        graphicobject.GraphicObjectLog.__init__(self, l)
        l.connect('point_changed', self, LeaderLog.pointChange)
        l.connect('size_changed', self, LeaderLog.sizeChange)

    def pointChange(self, l, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old endpoint: " + `_old`
        _oid = _old.getID()
        _new = args[1]
        if not isinstance(_new, point.Point):
            raise TypeError, "Invalid new endpoint: " + `_new`
        _nid = _new.getID()
        self.saveUndoData('point_changed', _oid, _nid)

    def sizeChange(self, l, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _size = args[0]
        if not isinstance(_size, float):
            _size = float(args[0])
        self.saveUndoData('size_changed', _size)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _l = self.getObject()
        _op = args[0]
        if _op == 'point_changed':
            if _alen < 3:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _nid = args[2]
            _p1, _p2, _p3 = _l.getPoints()
            if undo:
                self.saveRedoData('point_changed', _oid, _nid)
                self.ignore('point_changed')
                try:
                    _parent = _l.getParent()
                    if _parent is None:
                        raise ValueError, "Leader has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Old point missing: id=%d" % _oid
                    _x1, _y1 = _p1.getCoords()
                    _x2, _y2 = _p2.getCoords()
                    _x3, _y3 = _p3.getCoords()
                    _moved = False
                    _l.startUndo(True)
                    try:
                        if _p1.getID() == _nid:
                            _l.setP1(_op)
                            if (abs(_op.x - _x1) > 1e-10 or
                                abs(_op.y - _y1) > 1e-10):
                                _moved = True
                        elif _p2.getID() == _nid:
                            _l.setP2(_op)
                            if (abs(_op.x - _x2) > 1e-10 or
                                abs(_op.y - _y2) > 1e-10):
                                _moved = True
                        elif _p3.getID() == _nid:
                            _l.setP3(_op)
                            if (abs(_op.x - _x3) > 1e-10 or
                                abs(_op.y - _y3) > 1e-10):
                                _moved = True
                        else:
                            raise ValueError, "Unexpected point ID: %d" % _nid
                    finally:
                        _l.endUndo()
                    if _moved:
                        _l.sendMessage('moved', _x1, _y1, _x2, _y2, _x3, _y3)
                finally:
                    self.receive('point_changed')
            else:
                self.saveUndoData('point_changed', _oid, _nid)
                _parent = _l.getParent()
                if _parent is None:
                    raise ValueError, "Leader has no parent - cannot undo"
                _np = _parent.getObject(_nid)
                if _np is None or not isinstance(_np, point.Point):
                    raise ValueError, "New point missing: id=%d" % _nid
                self.ignore('point_changed')
                try:
                    if _p1.getID() == _oid:
                        _l.setP1(_np)
                    elif _p2.getID() == _oid:
                        _l.setP2(_np)
                    elif _p3.getID() == _oid:
                        _l.setP3(_np)
                    else:
                        raise ValueError, "Unexpected point ID: %d" % _oid
                finally:
                    self.receive('point_changed')
        elif _op == 'size_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _size = args[1]
            if not isinstance(_size, float):
                _size = float(args[1])
            if undo:
                self.saveUndoData('size_changed', _l.getArrowSize())
            else:
                self.saveRedoData('size_changed', _l.getArrowSize())
            self.ignore('size_changed')
            try:
                if undo:
                    _l.startUndo(True)
                    try:
                        _l.setArrowSize(_size)
                    finally:
                        _l.endUndo()
                else:
                    _l.setArrowSize(_size)
            finally:
                self.receive('size_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op




#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Subclass of NSOutlineView, used to diplay layers in image
#

import Globals
import Generic.layer
from AppKit import NSOutlineView, NSMenu, NSMenuItem

class LayerOutlineView(NSOutlineView):
    """ Subclass of NSOutlineView, used to display layers
    
The purpose of the subclass is primarily to allow custom
context menus for the layer items. 

    """
    def menuForEvent_(self, event):
        _point = self.convertPoint_fromView_(event.locationInWindow(), None)
        _index = self.rowAtPoint_(_point)
        if -1 == _index:
            return None

        if not self.isRowSelected_(_index):
            self.selectRow_byExtendingSelection_(_index, False)

        _item = self.itemAtRow_(_index)
        _layer = Globals.unwrap(_item)
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        
        _menu = NSMenu.alloc().initWithTitle_("Context")
        _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Rename", "EditLayerName:", "")
        _menu.addItem_(_menuItem)
        if _layer.isVisible():
            _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Hide", "HideLayer:", "")
        else:
            _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Show", "ShowLayer:", "")            
        _menuItem.setRepresentedObject_(_item)
        _menu.addItem_(_menuItem)
        _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Add Child Layer", "AddChildLayer:", "")
        _menuItem.setRepresentedObject_(_item)
        _menu.addItem_(_menuItem)
        _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Clear", "ClearLayer:", "")
        _menuItem.setRepresentedObject_(_item)
        _menu.addItem_(_menuItem)
        if _layer.hasChildren():
            _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Hide Children", "HideChildLayers:", "")
            _menuItem.setRepresentedObject_(_item)
            _menu.insertItem_atIndex_(_menuItem, 3)
            _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Show Children", "ShowChildLayers:", "")
            _menuItem.setRepresentedObject_(_item)
            _menu.insertItem_atIndex_(_menuItem, 4)
        elif _layer.getParentLayer() is not None:
            _menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Delete", "DeleteLayer:", "")
            _menuItem.setRepresentedObject_(_item)
            _menu.addItem_(_menuItem)
        return _menu
            
        
        


        
            
            


#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# this is a class that keeps a tolerance value. It should
# be used as a class (static) variable for classes that
# will compare instances with a tolerance
#
# a valid tolerance is a float value 0 or greater
#

TOL = 1e-10

class TolObject(object):
    """A class for maintaining a tolerance value.

The tolerance value is a float that must be 0.0 or
greater. Any class using this class as a base class
will have a tolerance value unique to that class.

There are two member functions:

getTolerance(): return the current tolerance
setTolerance(t): set the new tolerance

    """
    def __init__(self, t=None):
        """Initialize a TolObject.

TolObject(t)

Optional argument t must be a float, and it
must be greater than 0. A default tolerance is set
if the function is called without arguments.
        """
        if t is None:
            t = TOL
        tol = t
        if not isinstance(tol, float):
            tol = float(t)
        if tol < 0.0:
            raise ValueError, "Tolerance must be greater than 0: " + `tol`
        self.__tolerance = tol
        
    def setTolerance(self, t=None):
        """Set the tolerance value.

setTolerance(t)

Optional argument t must be a float, and it
must be greater than 0. The default tolerance is
reset if the function is called without arguments.
This function returns the old tolerance value.
        """
        old_tol = self.__tolerance
        if t is None:
            t = TOL
        tol = t
        if not isinstance(tol, float):
            tol = float(t)
        if tol < 0.0:
            raise ValueError, "Tolerance must be greater than 0: " + `tol`
        self.__tolerance = tol
        return old_tol

    def getTolerance(self):
        """Get the tolerance value.

getTolerance()

Return the current tolerance.
        """
        return self.__tolerance

    tolerance = property(getTolerance, setTolerance, None, "Tolerance value")

class StaticTolObject(object):
    """A class for maintaining a tolerance value.

This class is meant to be a base-class for classes
that wish to use a tolerance value for comparing
one instance to another.

There are two class methods:

getTolerance(): return the current tolerance
setTolerance(tol): set the new tolerance

This class stores the tolerance value as a static class
variable, so any classes using this class as a base class
will share the same tolerance value.
    """

    __tolerance = TOL
    
    def setTolerance(cls, t=None):
        """Set the tolerance value.

Optional argument t must be a float, and itmust be
greater than 0. The default tolerance is reset if
the function is called without arguments.

This function returns the old tolerance value.
        """
        old_tol = cls.__tolerance
        if t is None:
            t = TOL
        _t = t
        if not isinstance(_t, float):
            _t = float(t)
        if _t < 0.0:
            raise ValueError, "Tolerance must be greater than 0: " + `_t`
        cls.__tolerance = _t
        return old_tol

    setTolerance = classmethod(setTolerance)
    
    def getTolerance(cls):
        """Get the tolerance value.

Return the current tolerance.
        """
        return cls.__tolerance

    getTolerance = classmethod(getTolerance)

def toltest(tol):
    """Test that a tolerance value is valid.

toltest(tol)

The argument "tol" should be a float.
    """
    _t = tol
    if not isinstance(_t, float):
        _t = float(tol)
    if _t < TOL:
        raise ValueError, "Invalid tolerance: %g" % _t
    return _t



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Cocoa CAD drawing window controller, delegate of split view
#

import string
import Generic.layer
import Generic.keywords

import CADView
import LayerView
import CocoaPrompt
import Globals
import AppController
from Globals import WinFrameName, LayerDeletedNotification, LayerAddedNotification, ToolChangedNotification

from PyObjCTools import NibClassBuilder
from AppKit import *
from Foundation import NSNotificationCenter

#
# Define True & False in case this is Python < 2.2.1
#
try:
    True, False
except NameError:
    True, False = 1, 0

    
class ImageWindowController(NSWindowController, NSOutlineViewDataSource):
    """ Custom NSWindowController for ImageDocument window
    
An ImageWindowController is a NSWindowController used to manage
the NSWindow used by the ImageDocument class.  It is the delegate
of the NSSplitView & NSOutlineView (holding the layers), & NSWindow
    """
#
# Beginning of Cocoa methods
#

    def initWithWindowNibName_owner_(self, nib, owner):
        self = super(ImageWindowController, self).initWithWindowNibName_owner_(nib, owner)
        if (self):
            self.__outlineView = None
            self.__splitView = None
        return self
        
    def windowDidLoad(self):
        _doc = self.document()
        _sv = _doc.getSplitView()
        _ov = _doc.getOutlineView()
        _text = _doc.getEntry()
        _win = self.window()
        _text.setDelegate_(self)
        _sv.setDelegate_(self)
        _win.setDelegate_(self)
        _win.setFrameUsingName_(WinFrameName)
        _ov.setDataSource_(self)
        _ov.setDelegate_(self)
        _ov.sizeLastColumnToFit()
        _ov.setAutoresizesOutlineColumn_(True)
        _da = _doc.getDA()
        _scrv = _da.enclosingScrollView()
        _scrvSize = _scrv.contentSize()
        _scrv.setCopiesOnScroll_(False)
        _da.setFrameSize_(_scrvSize)
        #
        # ensure selected layer is active
        #
        _item = _ov.itemAtRow_(0)
        _ov.expandItem_expandChildren_(_item, True)
        _endRow = _ov.numberOfRows() - 1
        _ov.selectRow_byExtendingSelection_(_endRow, False)
        _ov.selectRow_byExtendingSelection_(0, False)
        #
        # Register for notifications of changes to the document
        #
        _img = _doc.getImage()
        _obj = Globals.wrap(_img)
        _nc = NSNotificationCenter.defaultCenter()
        _nc.addObserver_selector_name_object_(self, "layerChanged:", LayerAddedNotification, _obj)
        _nc.addObserver_selector_name_object_(self, "layerChanged:", LayerDeletedNotification, _obj)
        _nc.addObserver_selector_name_object_(self, "toolChanged:", ToolChangedNotification, _doc)
        
    def windowDidResize_(self, note):
        _win = note.object()
        _win.saveFrameUsingName_(WinFrameName)
        
    def windowWillClose_(self, note):
        _nc = NSNotificationCenter.defaultCenter()
        _nc.removeObserver_(self)
        
        
#
# Beginning of Split View delegate methods
#

    def splitView_resizeSubviewsWithOldSize_(self, splitView, oldSize):
        """Resize split view subviews after size change to oldSize

splitView_resizeSubviewsWithOldSize_(self, splitView, oldSize)
        """
        _ov = self.document().getOutlineView()
        _cadView = self.document().getDA()
        _subviewArray = splitView.subviews()

        for _count in range(_subviewArray.count()):
            _aView = _subviewArray.objectAtIndex_(_count)
            if (_ov.isDescendantOf_(_aView)):
                _ov = _aView
            elif (_cadView.isDescendantOf_(_aView)):
                _cadView = _aView
                    
        _ovFrame = _ov.frame()
        _cadFrame = _cadView.frame()
        _splitSize = splitView.frame().size
        _extraWidth = splitView.dividerThickness() + _ovFrame.size.width
        _ovFrame.size.height = _splitSize.height
        _cadFrame.size.height = _splitSize.height
        if _extraWidth < _splitSize.width:
            _cadFrame.size.width = _splitSize.width - _extraWidth 
        else:
            splitView.adjustSubviews()
        _ov.setFrame_(_ovFrame)
        _cadView.setFrame_(_cadFrame)
                
#
# End of Split View delegate methods
#

#
# Beginning of Outline View delegate methods
#
    def outlineViewSelectionDidChange_(self, note):
        _ov = note.object()
        _index = _ov.selectedRow()
        if -1 != _index:
            _item = _ov.itemAtRow_(_index)
            _layer = Globals.unwrap(_item)
            _doc = self.document()
            _doc.getImage().setActiveLayer(_layer)
            _doc.getDA().setNeedsDisplay_(True)
    
    def outlineViewItemDidCollapse_(self, note):
        _doc = self.document()
        _active_layer = _doc.getImage().getActiveLayer()
        _ov = note.object()
        _row = _ov.rowForItem_(Globals.wrap(_active_layer))
        if -1 == _row:
            _item = note.userInfo()["NSObject"]
            _row = _ov.rowForItem_(_item)
            _ov.selectRow_byExtendingSelection_(_row, False)
    
    def outlineView_willDisplayCell_forTableColumn_item_(self, ov, cell, tc, item):
        _layer = Globals.unwrap(item)
        if _layer.isVisible():
            _color = NSColor.blackColor()
        else:
            _color = NSColor.lightGrayColor()
        cell.setTextColor_(_color)  
        
#
# Outline view data source methods
#
#
# Beginning of OutlineView Data source methods
#
    def outlineView_isItemExpandable_(self, outlineView, item):
        """ NSOutlineViewDataSource method for layer view 

outlineView_isItemExpandable_(outlineView, item)
        """
        _item = Globals.unwrap(item)
        if _item is None:
            return True # Top Layer item is always displayed
        return _item.hasSublayers()
    
    def outlineView_numberOfChildrenOfItem_(self, outlineView, item):
        """ NSOutlineViewDataSource method for layer view 

outlineView_numberOfChildrenOfItem_(outlineView, item)
        """
        _item = Globals.unwrap(item)
        if _item is None:
            return 1 # Top Layer item is first item in layer list
        return len(_item.getSublayers())
        
    def outlineView_child_ofItem_(self, outlineView, itemIndex, item):
        """ NSOutlineViewDataSource method for layer view 

outlineView_child_ofItem_(outlineView, itemIndex, item)
        """
        _item = Globals.unwrap(item)
        if _item is None:
            return Globals.wrap(self.document().getImage().getTopLayer())
        _children = _item.getSublayers()
        return Globals.wrap(_children[itemIndex])
        
    def outlineView_objectValueForTableColumn_byItem_(self, ov, tc, item):
        """ NSOutlineViewDataSource method for layer view 

outlineView_objectValueForTableColumn_byItem_(outlineView, tableColumn, item)
        """
        if tc.identifier() != Globals.LayerColumnName:
            return None
        _item = Globals.unwrap(item)
        if _item is None:
            _item = self.document().getImage().getTopLayer()
        return _item.getName()
        
    def outlineView_setObjectValue_forTableColumn_byItem_(self, ov, val, tc, item):
        """ NSOutlineViewDataSource method for layer view
        
outlineView_setObjectValue_forTableColumn_byItem_(outlineView, value, tableColumn, item)
        """
        if tc.identifier() != Globals.LayerColumnName:
            return
        _layer = Globals.unwrap(item)
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid active Layer: " + `_layer`
        _layer.setName(val)


    def layerChanged_(self, note):
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        _dict = note.userInfo()
        _parent = Globals.wrap(_dict["parent"])
        _ov = _doc.getOutlineView()
        _image = _doc.getImage()
        _l = _activeLayer = _image.getActiveLayer()
#        _ov.reloadItem_reloadChildren_(_parent, True)
        _ov.reloadData()
        _topLayer = _image.getTopLayer()
        _stack = []
        while _l is not _topLayer:
            _l = _l.getParentLayer()
            _stack.append(_l)
        while len(_stack):
            _item = Globals.wrap(_stack.pop())
            _ov.expandItem_(_item)
        _al = Globals.wrap(_activeLayer)
        _row = _ov.rowForItem_(_al)
        _ov.selectRow_byExtendingSelection_(_row, False)
        
    def toolChanged_(self, note):
        _sheet = self.window().attachedSheet()
        if _sheet is None:
            return
        _userinfo = note.userInfo()
        _tool = _userinfo["tool"]
        if not isinstance(_tool, Generic.tools.TextTool):
            NSApplication.sharedApplication().endSheet_(_sheet)
            _sheet.orderOut_(_sheet)
            
        
               
#
# Layer context menu functions
#    
    def EditLayerName_(self, sender):
        _ov = self.document().getOutlineView()
        _row = _ov.selectedRow()
        _ov.editColumn_row_withEvent_select_(0, _row, None, True)
        
    def HideLayer_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _layer.hide()
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        _doc.getOutlineView().setNeedsDisplay_(True)

    def ShowLayer_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _layer.show()
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        _doc.getOutlineView().setNeedsDisplay_(True)
                
    def AddChildLayer_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _doc = self.document()
        _doc.addChildLayer(_layer)
        
    def HideChildLayers_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _children = _layer.getSublayers()
        while(len(_children)):
            _child = _children.pop()
            _child.hide()
            if _child.hasSublayers():
                _children = _children + _child.getSublayers()
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        _doc.getOutlineView().setNeedsDisplay_(True)
        
    def ShowChildLayers_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _children = _layer.getSublayers()
        while(len(_children)):
            _child = _children.pop()
            _child.show()
            if _child.hasSublayers():
                _children = _children + _child.getSublayers()
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        _doc.getOutlineView().setNeedsDisplay_(True)
        
    def ClearLayer_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _layer.clear()
        _doc = self.document()
        _doc.getDA().setNeedsDisplay_(True)
        
    def DeleteLayer_(self, sender):
        _layer = Globals.unwrap(sender.representedObject())
        if not isinstance(_layer, Generic.layer.Layer):
            raise TypeError, "Invalid Layer: " + `_layer`
        _doc = self.document()
        _doc.delLayer(_layer)

#
# End of Layer context menu functions       
#
       
#
#   somebody entered something in the entry event
#
    def controlTextDidEndEditing_(self, note):
        _doc = self.document()
        _tool = _doc.getTool()
        _field = note.object()
        _text = string.strip(string.lower(_field.stringValue()))
        _cmds = Generic.keywords.defaultglobals
        if _text == 'end' or _text == 'stop':
            _doc.reset()
        elif _text in _cmds:
                _opt = _cmds[_text]
                _cmd = CocoaPrompt.lookup(_opt)
                NSApp = NSApplication.sharedApplication()
                try:
                    eval(_cmd)
                except:
                    NSBeep()
        elif _tool is not None and _tool.hasHandler("entry_event"):
            _handler = _tool.getHandler("entry_event")
            try: 
                _handler(_doc, _text, _tool)
            except:
                NSBeep()

    

#
# Beginning of NSWindow delegate methods
#
#    def windowWillResize_toSize_(self, win, size):
#        if not self.__inResize:
#            self.__inResize = True
#            _da = self.document().getDA()
#            _img = _da.getImageRep()
#            _iv = NSImageView.alloc().initWithFrame_(_da.frame())
#            _iv.setImage_(_img)
#            self.__dasv = _da.enclosingScrollView() 
#            _da.removeFromSuperviewWithoutNeedingDisplay()
#            self.__dasv.setDocumentView_(_iv)
#        return size
        
        
        
#
# End of NSWindow Delegate methods
#

#
# End of Cocoa methods
#

    
